/*
    Copyright (c) 2013,2014 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc> 

    This file is part of dsid-tools.

    This file is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This file is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with This file. If not, see <http://www.gnu.org/licenses/>.

*/

#include "autoconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <errno.h>
#include <digitalSTROM/ds.h>
#include <digitalSTROM/dsuid/dsuid.h>

#if defined(HAVE_GETOPT_H) && defined(HAVE_GETOPT_LONG)
    #include <getopt.h>
    #define OPTSTR "i:m:uvhs"
#else
    #error Need getopt_long!
#endif

static void print_copyright()
{
    printf("\nmac2dsuid is part of dSID tools version %s\n", VERSION);
    printf("Copyright (c) 2013 aizo ag, Zurich, Switzerland\n\n");
}


static void print_usage(const char *myname)
{
    print_copyright();

    fprintf(stderr, "\n%s options\n\
\n\
Supported options:\n\
    --mac or -m         convert given mac address to dSUID\n\
    --interface or -i   convert mac of given interface (i.e. eth0) to dSUID\n\
    --vdsm              generate vdSM dSUID based on the mac address\n\
    --dsid              print a dSID instead of a dSUID\n\
    --version           print version information and exit\n\
    --help              this help message\n\
\n\
    Combination of -m and -i is not allowed, one of the options must be\
given.\n\
\n", myname);
}

static void stripcolons(char *mac)
{
    int i = 0;
    while (mac[i] != '\0')
    {
        if (mac[i] == ':')
        {
            int j = i;
            while (mac[j] != '\0')
            {
                mac[j] = mac[j + 1];
                j++;
            }
        }
        else
        {
            i++;
        }
    }
}

void print_dsuid(dsuid_t d)
{
    int i;
    for (i = 0; i < DSUID_SIZE; i++)
    {
        printf("%02x", (int)d.id[i]);
    }
    printf("\n");
}

int main(int argc, char **argv)
{
    int opt_index = 0;
    int o;
    int i;
    bool printDsid = false;
    bool vdsm = false;
    const char *interface = NULL;
    const char *mac = NULL;
    char macbuf[18];

    static struct option long_options[] =
    {
        {"mac",         1, 0, 'm'},
        {"interface",   1, 0, 'i'},
        {"vdsm",        0, 0, 's'},
        {"dsid",        0, 0, 'u'},
        {"version",     0, 0, 'v'},
        {"help",        0, 0, 'h'},
        {0, 0, 0, 0}
    };

    while (1)
    {
        o = getopt_long(argc, argv, OPTSTR, long_options, &opt_index);
        if (o == -1) break;

        switch(o)
        {
            case 'm':
                mac = optarg;
                break;
            case 'i':
                interface = optarg;
                break;
            case 'u':
                printDsid = true;
                break;
            case 's':
                vdsm = true;
                break;
            case 'v':
                print_copyright();
                exit(EXIT_SUCCESS);
            default:
                print_usage(argv[0]);
                exit(EXIT_FAILURE);
        }
    }

    if ((mac != NULL) && (interface != NULL))
    {
        print_usage(argv[0]);
        fprintf(stderr,
                "Error: can't use mac and interface options at the same "
                "time!\n");
        exit(EXIT_FAILURE);
    }

    if ((mac == NULL) && (interface == NULL))
    {
        print_usage(argv[0]);
        fprintf(stderr, 
                "Error: please provide either mac or interface option!\n");
        exit(EXIT_FAILURE);
    }

    if (printDsid && vdsm)
    {
        fprintf(stderr, "Error: "
                "only dSUID generation for vdSM is supported!\n");
    }

    if (interface != NULL)
    {
        char tmp[256] = "/sys/class/net/";
        tmp[255] = '\0';
        size_t len = strlen(tmp);
        size_t len_addr = strlen("/address");

        if (strlen(interface) > (sizeof(tmp) - 1 - len - len_addr))
        {
            fprintf(stderr, "Interface name too long!");
            exit(EXIT_FAILURE);
        }
       
        strncpy(tmp + len, interface, sizeof(tmp) - 1 - len);

        len = strlen(tmp);
        if (len < sizeof(tmp) - 1)
        {
            strncpy(tmp + len, "/address", sizeof(tmp) - 1 - len - len_addr);
        }

        FILE *f = fopen(tmp, "r");
        if (f == NULL)
        {
            fprintf(stderr, "Could not open %s: %s\n", tmp, strerror(errno));
            exit(EXIT_FAILURE);
        }

        size_t bytes_read = fread(macbuf, sizeof(char), sizeof(macbuf) - 1, f);
        if (bytes_read != sizeof(macbuf) - 1)
        {
            fclose(f);
            fprintf(stderr, "Could not read %s\n", tmp);
            exit(EXIT_FAILURE);
        }
        fclose(f);
    }

    if (mac != NULL)
    {
        strncpy(macbuf, mac, sizeof(macbuf) - 1);
    }

    macbuf[sizeof(macbuf)-1] = '\0';
    stripcolons(macbuf);

    dsuid_t d;

    if (vdsm)
    {
        int ret = dsuid_generate_v3_from_namespace(DSUID_NS_VDSM, macbuf, &d);
        if (ret != DSUID_RC_OK)
        {
            fprintf(stderr, "Failed to generate dSUID for vDSM: "
                    "libdsuid code %d\n", ret);
            exit(EXIT_FAILURE);
        }
        print_dsuid(d);
    }
    else
    {
        DsmApiGetEthernetDSUID(macbuf, &d);
        if (printDsid)
        {
            for (i = 0; i < 12; i++)
            {
                int j = (i < 6)? 0 : 4;
                printf("%02x", (int)d.id[i+j]);
            }
            printf("\n");
        } else {
            print_dsuid(d);
        }
    }

    return EXIT_SUCCESS;
}
