/*
    Copyright (c) 2009 digitalSTROM.org, Zurich, Switzerland

    This file was originally part of the dsplwriter project (GPLv3).

    Adaptions to dss11-ledd:
    Copyright (c) 2012 aizo ag, Zurich, Switzerland


    Author: Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc> 

    dss11-ledd is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dss11-ledd is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with leddwriter. If not, see <http://www.gnu.org/licenses/>.

*/

#include "config.h"

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include "ledd.h"
#include "sockutil.h"

#define PIPE_R 0
#define PIPE_W 1
#define CONNMAX 4

static int listen_sockfd = -1;
static int shutdown_pipe[2];
static int connected_fds[CONNMAX];
static int open_connections = 0;
static int ledd_port = 50501;
static ledd_action_callback handler_callback = NULL;
static void* argument = NULL;

static int qsort_intcompare(const void *v1, const void *v2)
{
    const int *i1 = (const int *) v1;
    const int *i2 = (const int *) v2;

    return (*i1 < *i2) - (*i1 > *i2);
};

ledd_status ledd_init(int port, ledd_action_callback callback, void *arg)
{
    struct sockaddr_in addr;
    int reuse_on = 1;
    int retcode;
    int i;

    ledd_port = port;
    handler_callback = callback;
    argument = arg;
    struct linger
    {
        int l_on;
        int l_linger;
    } linger;

    linger.l_on = 1;
    linger.l_linger = 30;

    listen_sockfd = socket(AF_INET, SOCK_STREAM, 0);
    if (listen_sockfd < 0)
    {
        fprintf(stderr, "Failed to create socket: %s\n", strerror(errno));
        return LEDD_ERROR;
    }

    retcode = setsockopt(listen_sockfd, SOL_SOCKET, SO_REUSEADDR, 
                         (const char*)&reuse_on, sizeof(int));
    if (retcode < 0)
    {
        fprintf(stderr, "Failed to set socket option SO_REUSEDDR: %s\n",
                strerror(errno));
        return LEDD_ERROR;
    }

    retcode = setsockopt(listen_sockfd, SOL_SOCKET, SO_LINGER, 
                         (const char*)&linger, sizeof(linger));
    if (retcode < 0)
    {
        fprintf(stderr, "Failed to set socket option SO_LINGER: %s\n",
                strerror(errno));
        return LEDD_ERROR;
    }
    
    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;
    addr.sin_port = htons(port);

    retcode = bind(listen_sockfd,  (struct sockaddr *)&addr, sizeof(addr));
    if (retcode < 0)
    {
        fprintf(stderr, "Failed to bind: %s\n", strerror(errno));
        return LEDD_ERROR;
    }

    shutdown_pipe[PIPE_R] = -1;
    shutdown_pipe[PIPE_W] = -1;

    if (pipe(shutdown_pipe) < 0)
    {
        fprintf(stderr, "Failed to create pipe: %s\n", strerror(errno));
        return LEDD_ERROR;
    }

    retcode = listen(listen_sockfd, CONNMAX);
    if (retcode < 0)
    {
        fprintf(stderr, "Failed to listen on socket: %s\n", strerror(errno));
        return LEDD_ERROR;
    }

    for (i = 0; i < CONNMAX; i++)
        connected_fds[i] = -1;

    return LEDD_OK;
}

void ledd_cleanup()
{
    int i;

    if (listen_sockfd > -1)
    {
        shutdown(listen_sockfd, SHUT_RDWR);
        listen_sockfd = -1;
    }

    if (shutdown_pipe[PIPE_R] > -1)
    {
        close(shutdown_pipe[PIPE_R]);
        shutdown_pipe[PIPE_R] = -1;
    }

    if (shutdown_pipe[PIPE_W] > -1)
    {
        close(shutdown_pipe[PIPE_W]);
        shutdown_pipe[PIPE_W] = -1;
    }

    for (i = 0; i < open_connections; i++)
    {
        if (connected_fds[i] > -1)
        {
            shutdown(connected_fds[i], SHUT_RDWR);
            connected_fds[i] = -1;
        }
    }

    open_connections = 0;
}

ledd_status ledd_process_event()
{
    fd_set rfds;
    int max_fd = 0;
    int i;
    int retcode;
    struct sockaddr fsaddr;
    socklen_t fromlen;
    int tmp;

    FD_ZERO(&rfds);
    
    max_fd = (listen_sockfd > max_fd ? listen_sockfd : max_fd);
    FD_SET(listen_sockfd, &rfds);

    for (i = 0; i < open_connections; i++)
    {
        max_fd = (connected_fds[i] > max_fd ? connected_fds[i] : max_fd);
        FD_SET(connected_fds[i], &rfds);
    }

    retcode = select(max_fd + 1, &rfds, NULL, NULL, NULL);
    if (retcode < 0)
    {
        if (errno == EINTR)
            return LEDD_OK;

        fprintf(stderr, "Select failed, aborting: %s\n", strerror(errno));
        return LEDD_ERROR;
    }

    if (retcode == 0)
        return LEDD_OK;

    if (FD_ISSET(shutdown_pipe[PIPE_R], &rfds))
    {
        struct ledd_event evt;
        memset(&evt, 0, sizeof(evt));
        read(shutdown_pipe[PIPE_R], &evt, sizeof(evt));
        if (evt.action == ledd_quit)
            return LEDD_QUIT;
    }

    if (FD_ISSET(listen_sockfd, &rfds))
    {
        if (open_connections < CONNMAX)
        {
            int new_fd = accept(listen_sockfd, &fsaddr, &fromlen);
            if (new_fd < 0)
            {
                fprintf(stderr, "Could not accept connection: %s\n", 
                        strerror(errno));
                // this is not a fatal failure
                return LEDD_OK;
            }

            open_connections++;
            connected_fds[open_connections-1] = new_fd;
        }
        else
        {
            int reject_fd = accept(listen_sockfd, &fsaddr, &fromlen);
            if (reject_fd >= 0)
            {
                fprintf(stderr, "Maximum connections reached, rejecting " 
                                "incoming connection!\n");
                shutdown(reject_fd, SHUT_RDWR);
            }
        }
    }

    tmp = open_connections;
    for (i = 0; i < open_connections; i++)
    {
        if (FD_ISSET(connected_fds[i], &rfds))
        {
            struct ledd_event evt;
            memset(&evt, 0, sizeof(struct ledd_event));
            ssize_t len = sockread(connected_fds[i], (unsigned char*)&evt, 
                          sizeof(struct ledd_event), 2);

            if (len <= 0)
            {
                shutdown(connected_fds[i], SHUT_RDWR);
                connected_fds[i] = -1;
                tmp--;
                if (tmp < 0)
                    tmp = 0;

                continue;
            }

            if ((evt.action < ledd_quit) || (evt.action >= ledd_unhandled))
            {
                continue;
            }
            if (handler_callback != NULL) {
                handler_callback(evt.action, argument);
            }
        }
    }

    open_connections = tmp;
    qsort(connected_fds, CONNMAX, sizeof(int), qsort_intcompare);

    return LEDD_OK;
}

void ledd_send_quit()
{
    struct ledd_event evt;
    evt.action = ledd_quit;
    write(shutdown_pipe[PIPE_W], (unsigned char*)&evt, sizeof(struct ledd_event));
}

