/*
 * dSS11 led handling daemon
 * Copyright (c) 2011,2012 aizo ag, Switzerland
 *
 * Authors: Johannes Winkelmann <johannes.winkelmann@aizo.com>
 *          Michael Tross <michael.tross@aizo.com>
 *          Sergey 'Jin' Bostandzhyan <jin@mediatomb.cc>
 */

#include <stdint.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

#include <cstdlib>
#include <string>
#include <iostream>
#include <vector>
#include <utility>
#include <fstream>
#include <sstream>

#include "config.h"
#include "led.h"
#include "ledgroup.h"
#include "sequence.h"
#include "commandhandler.h"
#include "report.h"
#include "ledd.h"

extern char *optarg;
extern int optind;

int debugLevel = 7;

/* process id file */
static char* pidfile = (char *) "/var/run/dss11-ledd.pid";

/* running as daemon background process */
static int inBackground = 0;
static pthread_t main_thread_id;
static int shutdown_flag = 0;

static void signal_handler(int signum)
{
    if (pthread_self() != main_thread_id)
    {
        return;
    }

    if ((signum == SIGINT) || (signum == SIGTERM))
    {
        shutdown_flag++;
        if (shutdown_flag == 1)
        {
            fprintf(stderr, "Shutting down, please wait...\n");
            ledd_send_quit();
        }
        else if (shutdown_flag == 2)
        {
            fprintf(stderr, "Still shutting down, signal again to kill...\n");
        }
        else if (shutdown_flag > 2)
        {
            fprintf(stderr, "Clean shutdown failed, killing daemon...\n");
            exit(EXIT_FAILURE);
        }
    }
}

static void print_usage(void)
{
    fprintf(stderr, "Usage: dss11-ledd [OPTIONS]\n");
    fprintf(stderr, "       OPTION is a selection out of the following options:\n");
    fprintf(stderr, "       -h  --  print this help\n");
    fprintf(stderr, "       -d  --  debug-level: max debug level still logged: 0-7 (default: 5)\n");
    fprintf(stderr, "       -p  --  PID-file (default: /var/run/dss11-ledd.pid)\n");
    fprintf(stderr, "       -g  --  run in background\n");
    fprintf(stderr, "               log-output: foreground: stdout\n");
    fprintf(stderr, "                           daemonized: syslog\n");
    fprintf(stderr, "\n");
}

static int daemonize(void)
{
  int fd;
  pid_t pid;

  switch (pid = fork()) {
  case -1:
    return -1;
  case 0:        /* child side */
	break;
  default:    /* parent side */
	exit(0);
  }

  if (setsid() == -1)
	return -1;
  if (chdir("/") == -1)
	return -1;

  if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
	dup2(fd, STDIN_FILENO);
	dup2(fd, STDOUT_FILENO);
	dup2(fd, STDERR_FILENO);
	if (fd > 2)
	  close(fd);
  }
  inBackground = 1;
  return 0;
}


class AsciiLED
  : public LED
{
public:
  AsciiLED(const std::string& name, uint8_t state)
    : LED(state),
      m_name(name) {}

private:
  virtual void turnOn()  { std::cout << m_name << ": on"  << std::endl; }
  virtual void turnOff() { std::cout << m_name << ": off" << std::endl; }

  std::string m_name;
};

class GpioLED
  : public LED
{
public:
  GpioLED(const std::string& name, uint8_t state)
    : LED(state),
      m_path("/sys/class/leds/" + name + "/brightness")
    {}

private:
  virtual void turnOn()  {
    writeVal("1");
  }
  virtual void turnOff() {
    writeVal("0");
  }
  virtual void writeVal(const std::string& s)  {
    std::ofstream f(m_path.c_str());
    if (f) {
      f << s << std::endl;
      f.close();
    } else {
      std::cerr << "Error while writing to " << m_path << std::endl;
    }
  }

  std::string m_path;
};

class GpioLEDFactory
  : public LedFactory
{
public:
  LED* createLED(const std::string& name, uint8_t state)
    { return new GpioLED(name, state); }
};

class AsciiLEDFactory
  : public LedFactory
{
public:
  LED* createLED(const std::string& name, uint8_t state)
    { return new AsciiLED(name, state); }
};


int main(int argc, char** argv)
{
  int c;
  int goBackground = 0;
  struct sigaction action;

  /* Look for options */
  while ((c = getopt(argc, argv, "hd:p:g")) != -1) {
    switch (c) {
    case 'h':
      print_usage();
      exit(0);
    case 'd':
      debugLevel = atoi(optarg);
	  break;
    case 'p':
      pidfile = strdup(optarg);
      break;
    case 'g':
      goBackground = 1;
      break;
    }
  }

  if (goBackground) {
      openlog("dss11-ledd", LOG_PID, LOG_USER);
      daemonize();
  }

  report(LOG_NOTICE, "%s, %s (built %s, %s)\n", PACKAGE_NAME, PACKAGE_VERSION, __DATE__, __TIME__);

  if (pidfile) {
    FILE* fp;
    if ((fp = fopen(pidfile, "w")) != NULL) {
      fprintf(fp, "%u\n", (unsigned int) getpid());
      fclose(fp);
    } else {
      report(LOG_ERR, "Cannot create PID file: %s\n", pidfile);
    }
  }

  LedFactory* ledFactory;

#ifdef TESTMODE
  ledFactory = new AsciiLEDFactory;
#else
  ledFactory = new GpioLEDFactory;
#endif

  // keep the command state independent from connection
  CommandHandler handler(ledFactory);

  main_thread_id = pthread_self();
  memset(&action, 0, sizeof(action));
  action.sa_handler = signal_handler;
  action.sa_flags = 0;
  sigfillset(&action.sa_mask);

  if (sigaction(SIGINT, &action, NULL) < 0)
  {
      fprintf(stderr, "Could not register SIGINT handler!\n");
  }

  if (sigaction(SIGTERM, &action, NULL) < 0)
  {
      fprintf(stderr, "Could not register SIGTERM handler!\n");
  }

  // register and ignore SIGPIPE
  if (sigaction(SIGPIPE, &action, NULL) < 0)
  {
      fprintf(stderr, "Could not register SIGPIPE handler!\n");
  }

  if (ledd_init(DSS11_LEDD_PORT, CommandHandler::handlerCallback,
                (void*)&handler) < 0)
  {
      exit(EXIT_FAILURE);
  }

  while (!shutdown_flag) {
    ledd_status ret = ledd_process_event();
    if (ret == LEDD_QUIT) {
      break;
    }
  }

  ledd_cleanup();

  return 0;
}
