/*
    Copyright (c) 2013 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of libcommchannel.

    libcommchannel is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    libcommchannel is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libcommchannel. If not, see <http://www.gnu.org/licenses/>.
*/

#include "autoconfig.h"

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <string>
#include <pthread.h>
#include <vector>
#include <utility>

#include "commchannel_int.h"
#include "exceptions_int.h"
#include "sockutil.h"
#include "util.h"

#define SELECT_TIMEOUT_SECONDS      5
#define MAX_DATA_SIZE               16384 // 16k
#define MAX_REPLY_TIMEOUT_SECONDS   60 // clean up socket if we get no reply

namespace CC {

CommunicationDataInternal::CommunicationDataInternal(CommunicationDataID_t id,
                                     unsigned char *data, uint32_t size) :
                                     m_id(id), m_data(data), m_size(size)
{
}

const CommunicationDataID_t CommunicationDataInternal::getID() const
{
    return m_id;
}

const unsigned char *CommunicationDataInternal::getDataPtr() const
{
    return m_data;
}

uint32_t CommunicationDataInternal::getDataSize() const
{
    return m_size;
}

CommunicationDataInternal::~CommunicationDataInternal()
{
    if (m_data)
    {
        free(m_data);
    }
}

CommunicationChannelServerInternal::Connection::Connection()
{
    fd = -1;
    replied = false;
    received = false;
    size = 0;
    data = NULL;
    try
    {
        timeutil::getTimespecAfterMillis(MAX_REPLY_TIMEOUT_SECONDS * 1000, &ts);
    }
    catch (Exception &ex)
    {
        ts.tv_sec = 0;
        ts.tv_nsec = 0;
    }
}

CommunicationChannelServerInternal::CommunicationChannelServerInternal(int port, int maxconn) :
                                                    m_port(port),
                                                    m_max_connections(maxconn)
{
    struct sockaddr_in addr;
    int reuse_on = 1;
    int retcode;

    m_stop_flag = false;
    m_socket_fd = -1;

    struct linger
    {
        int l_on;
        int l_linger;
    } linger;

    linger.l_on = 1;
    linger.l_linger = 30;

    m_socket_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (m_socket_fd < 0)
    {
        throw _CCException("Failed to create socket: " +
                         std::string(strerror(errno)));
    }

    retcode = setsockopt(m_socket_fd, SOL_SOCKET, SO_REUSEADDR,
                         (const char*)&reuse_on, sizeof(int));
    if (retcode < 0)
    {
        close(m_socket_fd);
        m_socket_fd = -1;

        throw _CCException("Failed to set socket option SO_REUSEDDR: " +
                std::string(strerror(errno)));
    }

    retcode = setsockopt(m_socket_fd, SOL_SOCKET, SO_LINGER,
                         (const char*)&linger, sizeof(linger));
    if (retcode < 0)
    {
        close(m_socket_fd);
        m_socket_fd = -1;

        throw _CCException("Failed to set socket option SO_LINGER: " +
                std::string(strerror(errno)));
    }

    memset(&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = inet_addr("127.0.0.1");
    addr.sin_port = htons(m_port);

    retcode = bind(m_socket_fd,  (struct sockaddr *)&addr, sizeof(addr));
    if (retcode < 0)
    {
        close(m_socket_fd);
        m_socket_fd = -1;

        throw _CCException("Failed to bind: " + std::string(strerror(errno)));
    }

    retcode = listen(m_socket_fd, m_max_connections);
    if (retcode < 0)
    {
        close(m_socket_fd);
        m_socket_fd = -1;
        throw _CCException("Failed to listen on socket: " +
                         std::string(strerror(errno)));
    }

    retcode = pthread_create(&m_listener_thread, NULL,
                             CommunicationChannelServerInternal::staticThreadProc, this);
    if (retcode != 0)
    {
        close(m_socket_fd);
        m_socket_fd = -1;
    }

}

CommunicationChannelServerInternal::~CommunicationChannelServerInternal()
{
    m_stop_flag = true;
    if (m_listener_thread)
    {
        pthread_join(m_listener_thread, NULL);
    }

    if (m_socket_fd > -1)
    {
        close(m_socket_fd);
        m_socket_fd = -1;
    }

    while (!m_connected_fds.empty())
    {
        boost::shared_ptr<Connection> c = m_connected_fds.front();
        if (c->fd > -1)
        {
            close(c->fd);
        }
        if (c->data != NULL)
        {
            free(c->data);
        }
        m_connected_fds.pop_front();
    }
}

void *CommunicationChannelServerInternal::staticThreadProc(void *arg)
{
    CommunicationChannelServerInternal* instance = (CommunicationChannelServerInternal *)arg;
    instance->listenerThread();
    pthread_exit(NULL);
    return NULL;
}

void CommunicationChannelServerInternal::listenerThread()
{
    fd_set rfds;
    int max_fd = 0;
    int retcode;
    struct sockaddr fsaddr;
    socklen_t fromlen;
    memset(&fsaddr, 0, sizeof(struct sockaddr));
    memset(&fromlen, 0, sizeof(socklen_t));
    struct timeval timeout;
    std::list<boost::shared_ptr<CommunicationData> > notify;

    while (!m_stop_flag)
    {
        max_fd = 0;
        notify.clear();
        timeout.tv_sec = SELECT_TIMEOUT_SECONDS;
        timeout.tv_usec = 0;

        FD_ZERO(&rfds);

        max_fd = (m_socket_fd > max_fd ? m_socket_fd : max_fd);
        FD_SET(m_socket_fd, &rfds);

        m_connections_mutex.lock();

        std::list<boost::shared_ptr<Connection> >::iterator i =
                                                    m_connected_fds.begin();
        struct timespec tsnow;
        timeutil::getTimespecNow(&tsnow);
        while (i != m_connected_fds.end())
        {
            struct timespec ts = (*i)->ts;
            if (((*i)->replied == true) ||
                (timeutil::compareTimespecs(&ts, &tsnow) > 0))
            {
                if ((*i)->fd != -1)
                {
                    close((*i)->fd);
                }

                if ((*i)->data != NULL)
                {
                    free((*i)->data);
                }
                i = m_connected_fds.erase(i);
                continue;
            }
            // don't select on this fd
            if ((*i)->received == true)
            {
                i++;
                continue;
            }
            max_fd = ((*i)->fd > max_fd ? (*i)->fd : max_fd);
            FD_SET((*i)->fd, &rfds);
            i++;
        }

        m_connections_mutex.unlock();

        retcode = select(max_fd + 1, &rfds, NULL, NULL, &timeout);
        if (retcode == 0)
        {
            // timeout - nothing to read, nothing to write
            continue;
        }

        if (retcode < 0)
        {
            if (errno == EINTR)
            {
                continue;
            }

            break;
        }

        m_connections_mutex.lock();
        if (FD_ISSET(m_socket_fd, &rfds))
        {
            if (m_connected_fds.size() < m_max_connections)
            {
                int new_fd = accept(m_socket_fd, &fsaddr, &fromlen);
                if (new_fd < 0)
                {
#ifndef NDEBUG
                    fprintf(stderr, "Could not accept connection: %s\n",
                            strerror(errno));
#endif
                    // this is not a fatal failure
                }
                else
                {
                    boost::shared_ptr<Connection> conn(new Connection());
                    conn->fd = new_fd;
                    m_connected_fds.push_back(conn);
                }
            }
            else
            {
                int reject_fd = accept(m_socket_fd, &fsaddr, &fromlen);
                if (reject_fd >= 0)
                {
#ifndef NDEBUG
                    fprintf(stderr, "Maximum connections reached, rejecting "
                                    "incoming connection! %d\n", reject_fd);
#endif
                    close(reject_fd);
                }
            }
        }

        i = m_connected_fds.begin();
        while (i != m_connected_fds.end())
        {
            if (((*i)->fd != -1) && (!(*i)->received) &&
                (FD_ISSET((*i)->fd, &rfds)))
            {
                uint32_t size;
                ssize_t len = sockread((*i)->fd,
                                    (unsigned char *)(&size),
                                     sizeof(uint32_t), SELECT_TIMEOUT_SECONDS);
                // if data packet length could not be read, then reject it
                if (len != sizeof(uint32_t))
                {
                    close((*i)->fd);
                    i = m_connected_fds.erase(i);
                    continue;
                }

                (*i)->size = ntohl(size);
                if ((*i)->size > MAX_DATA_SIZE)
                {
                    close((*i)->fd);
                    i = m_connected_fds.erase(i);
                    continue;
                }

                if ((*i)->size > 0)
                {
                    (*i)->data = (unsigned char *)malloc((*i)->size);
                    if ((*i)->data == NULL)
                    {
                        close((*i)->fd);
                        i = m_connected_fds.erase(i);
                        continue;
                    }

                    len = sockread((*i)->fd,
                                   (*i)->data,
                                   (*i)->size,
                                   SELECT_TIMEOUT_SECONDS);
                    if ((len < 0) || ((uint32_t)len != (*i)->size))
                    {
                        close((*i)->fd);
                        free((*i)->data);
                        i = m_connected_fds.erase(i);
                        continue;
                    }
                    // do not try to read data from that connection anymore
                    (*i)->received = true;
                    CommunicationDataID_t id;
                    id.fd = (*i)->fd;
                    id.ts = (*i)->ts;
                    boost::shared_ptr<CommunicationData> cbdata(
                        new CommunicationData(id, (*i)->data, (*i)->size));
                    (*i)->data = NULL;
                    notify.push_back(cbdata);
                }
            }
            i++;
        }

        m_connections_mutex.unlock();
        while (!notify.empty())
        {
            boost::shared_ptr<CommunicationData> c = notify.front();
            this->notifySubscribers(notify.front());
            notify.pop_front();
        }
    }
}

void CommunicationChannelServerInternal::sendReply(CommunicationDataID_t id,
                                           unsigned char *data,
                                           uint32_t length)
{
    boost::mutex::scoped_lock lock(m_connections_mutex);
    // check if the id is valid
    std::list<boost::shared_ptr<Connection> >::iterator i =
                                                    m_connected_fds.begin();
    boost::shared_ptr<Connection> c;

    while (i != m_connected_fds.end())
    {
        if ((id.fd == (*i)->fd) &&
            (timeutil::compareTimespecs(&(id.ts), &((*i)->ts)) == 0))
        {
            c = (*i);
            break;
        }
        i++;
    }

    if (c == NULL)
    {
        throw _CCException("Socket not found\n");
    }

    uint32_t netlen = htonl(length);
    ssize_t sent = sockwrite(c->fd,
                             (unsigned char *)&netlen,
                             sizeof(netlen), SELECT_TIMEOUT_SECONDS);
    if (sent != sizeof(netlen))
    {
        // socket can be cleaned up
        c->replied = true;
        throw _CCException("Failed to send data");
    }

    sent = sockwrite(c->fd, data, length, SELECT_TIMEOUT_SECONDS);
    if ((sent < 0) || ((uint32_t)sent != length))
    {
        // socket can be cleaned up
        c->replied = true;
        throw _CCException("Failed to send data");
    }

    // socket can be cleaned up
    c->replied = true;
}

void CommunicationChannelServerInternal::addCallback(CommunicationChannelCallback *cb)
{
    this->removeCallback(cb); // prevent double entries
    boost::mutex::scoped_lock lock(m_callback_list_mutex);
    m_callbacks.push_back(cb);
}

void CommunicationChannelServerInternal::removeCallback(CommunicationChannelCallback *cb)
{
    boost::mutex::scoped_lock lock(m_callback_list_mutex);
    std::list<CommunicationChannelCallback *>::iterator i = m_callbacks.begin();

    while (i != m_callbacks.end())
    {
        if ((*i) == cb)
        {
            i = m_callbacks.erase(i);
            break;
        }
        i++;
    }
}

void CommunicationChannelServerInternal::notifySubscribers(boost::shared_ptr<CommunicationData> data)
{
    boost::mutex::scoped_lock lock(m_callback_list_mutex);
    std::list<CommunicationChannelCallback *>::iterator i = m_callbacks.begin();

    while (i != m_callbacks.end())
    {
        // if we detect a NULL subscriber, get rid of it
        if ((*i) == NULL)
        {
            i = m_callbacks.erase(i);
        }
        else
        {
            try
            {
                (*i)->messageReceived(data);
            }
            catch (Exception &ex)
            {
#ifndef NDEBUG
                fprintf(stderr, "Exception caught in callback\n");
#endif
            }
            i++;
        }
    }
}

CommunicationChannelClientInternal::CommunicationChannelClientInternal(int port)
{
    m_port = port;
}

CommunicationChannelClientInternal::~CommunicationChannelClientInternal()
{
}

boost::shared_ptr<CommunicationData> CommunicationChannelClientInternal::sendData(
        unsigned char *data, uint32_t size)
{
    int socket_fd;
    int ret;

    struct sockaddr_in destination;
    memset(&destination, 0, sizeof(destination));
    destination.sin_family = AF_INET;
    destination.sin_addr.s_addr = inet_addr("127.0.0.1");
    destination.sin_port = htons(m_port);
    socket_fd = socket(AF_INET, SOCK_STREAM, 0);
    if (socket_fd < 0)
    {
        throw _CCException("Failed to create socket: " +
                         std::string(strerror(errno)));
    }

    ret = connect(socket_fd, (struct sockaddr *)&destination,
                  sizeof(destination));
    if (ret < 0)
    {
        close(socket_fd);
        throw _CCException("Failed to connect: " +
                         std::string(strerror(errno)));
    }
    uint32_t netlen = htonl(size);
    ssize_t sent = sockwrite(socket_fd, (unsigned char *)&netlen,
                             sizeof(netlen), SELECT_TIMEOUT_SECONDS);
    if (sent != sizeof(netlen))
    {
        close(socket_fd);
        throw _CCException("Failed to send data");
    }

    sent = sockwrite(socket_fd, data, size, SELECT_TIMEOUT_SECONDS);
    if ((sent < 0) || ((uint32_t)sent != size))
    {
        close(socket_fd);
        throw _CCException("Failed to send data");
    }

    unsigned char *repdata = NULL;
    uint32_t replylen;
    ssize_t received = sockread(socket_fd, (unsigned char *)(&replylen),
                         sizeof(uint32_t), SELECT_TIMEOUT_SECONDS);
    // if data packet length could not be read, then reject it
    if (received != sizeof(uint32_t))
    {
        close(socket_fd);
        throw _CCException("Failed to receive reply");
    }

    replylen = ntohl(replylen);
    if (replylen > MAX_DATA_SIZE)
    {
        close(socket_fd);
        throw _CCException("Rejecting too big response message");
    }

    if (replylen > 0)
    {
        repdata = (unsigned char *)malloc(replylen);
        if (repdata == NULL)
        {
            close(socket_fd);
            throw _CCException("Could not allocate memory for message");
        }

        received = sockread(socket_fd, repdata, replylen,
                            SELECT_TIMEOUT_SECONDS);
        if ((received < 0) || ((uint32_t)received != replylen))
        {
            close(socket_fd);
            free(repdata);
            repdata = NULL;
            throw _CCException("Could not read reply data");
        }
    }
    close(socket_fd);
    CommunicationDataID_t id;
    id.fd = -1;
    boost::shared_ptr<CommunicationData> out(new CommunicationData(id, repdata,
                                                                    replylen));
    repdata = NULL;
    return out;
}

} // namespace
